"""
Certificate data class
"""

from dataclasses import dataclass, asdict
from typing import Union, Literal
from datetime import datetime

from .Asset import Asset
from ...base.base_list import BaseList


@dataclass
class Certificate:
    """
    Represents a certificate in CERT.
    """

    id: int = None
    certhash: str = None
    keySize: int = None
    serialNumber: str = None
    validToDate: Union[str, datetime] = None
    validTo: Union[int, datetime] = None
    validFromDate: Union[str, datetime] = None
    validFrom: Union[int, datetime] = None
    signatureAlgorithm: str = None
    extendedValidation: bool = None
    createdDate: Union[str, datetime] = None
    dn: str = None
    subject: dict = None
    # subject is parsed into below fields:
    subject_organization: str = None
    subject_locality: str = None
    subject_name: str = None
    subject_country: str = None
    subject_organizationUnit: BaseList[str] = None
    # end of subject fields
    updateDate: Union[str, datetime] = None
    lastFound: Union[int, datetime] = None
    imported: bool = None
    selfSigned: bool = None
    issuer: dict = None
    # issuer is parsed into below fields:
    issuer_organization: str = None
    issuer_organizationUnit: BaseList[str] = None
    issuer_name: str = None
    issuer_country: str = None
    issuer_state: str = None
    issuer_certhash: str = None
    issuer_locality: str = None
    # end of issuer fields
    issuerCategory: str = None
    instanceCount: int = None
    assetCount: int = None
    sources: Union[list[str], BaseList[str]] = None
    assets: Union[list[dict], BaseList[object]] = None
    type: Literal["Leaf", "Intermediate", "Root"] = None
    rootissuer: dict = None
    # rootissuer is parsed into below fields:
    rootissuer_organization: str = None
    rootissuer_organizationUnit: BaseList[str] = None
    rootissuer_name: str = None
    rootissuer_country: str = None
    rootissuer_state: str = None
    rootissuer_certhash: str = None
    rootissuer_locality: str = None
    # end of rootissuer fields

    def __post_init__(self):
        INT_TO_DT_FIELDS = ["validTo", "validFrom", "lastFound"]
        STR_TO_DT_FIELDS = ["validToDate", "validFromDate", "createdDate", "updateDate"]
        DT_FIELDS = INT_TO_DT_FIELDS + STR_TO_DT_FIELDS

        for field in DT_FIELDS:
            if getattr(self, field):
                if field in INT_TO_DT_FIELDS and not isinstance(
                    getattr(self, field), datetime
                ):
                    setattr(
                        self, field, datetime.fromtimestamp(getattr(self, field) / 1000)
                    )
                elif field in STR_TO_DT_FIELDS and not isinstance(
                    getattr(self, field), datetime
                ):
                    setattr(self, field, datetime.fromisoformat(getattr(self, field)))
                else:
                    raise ValueError(f"Field {field} is not a valid datetime field.")

        if self.subject:
            setattr(self, "subject_organization", self.subject.get("organization"))
            setattr(self, "subject_locality", self.subject.get("locality"))
            setattr(self, "subject_name", self.subject.get("name"))
            setattr(self, "subject_country", self.subject.get("country"))
            setattr(
                self,
                "subject_organizationUnit",
                BaseList(self.subject.get("organizationUnit", [])),
            )
            setattr(self, "subject", None)

        if self.issuer:
            setattr(self, "issuer_organization", self.issuer.get("organization"))
            setattr(
                self,
                "issuer_organizationUnit",
                BaseList(self.issuer.get("organizationUnit", [])),
            )
            setattr(self, "issuer_name", self.issuer.get("name"))
            setattr(self, "issuer_country", self.issuer.get("country"))
            setattr(self, "issuer_state", self.issuer.get("state"))
            setattr(self, "issuer_certhash", self.issuer.get("certhash"))
            setattr(self, "issuer_locality", self.issuer.get("locality"))
            setattr(self, "issuer", None)

        if self.rootissuer:
            setattr(
                self, "rootissuer_organization", self.rootissuer.get("organization")
            )
            setattr(
                self,
                "rootissuer_organizationUnit",
                BaseList(self.rootissuer.get("organizationUnit", [])),
            )
            setattr(self, "rootissuer_name", self.rootissuer.get("name"))
            setattr(self, "rootissuer_country", self.rootissuer.get("country"))
            setattr(self, "rootissuer_state", self.rootissuer.get("state"))
            setattr(self, "rootissuer_certhash", self.rootissuer.get("certhash"))
            setattr(self, "rootissuer_locality", self.rootissuer.get("locality"))
            setattr(self, "rootissuer", None)

        if self.sources:
            setattr(self, "sources", BaseList(self.sources))

        if self.assets:
            setattr(self, "assets", BaseList([Asset(**x) for x in self.assets]))

    def to_dict(self) -> dict:
        """
        Returns the certificate as a dictionary.
        """
        return asdict(self)

    def __dict__(self) -> dict:
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        Creates a Certificate object from a dictionary.
        """
        return cls(**data)



"""
Global AssetView call schema
"""

from frozendict import frozendict

GAV_SCHEMA = frozendict(
    {
        "gav": {
            "url_type": "gateway",
            "count_assets": {
                "endpoint": "/am/v1/assets/host/count",
                "method": ["POST"],
                "valid_params": ["filter", "lastSeenAssetId", "lastModifiedDate"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "token",
            },
            "get_all_assets": {
                "endpoint": "/am/v1/assets/host/list",
                "method": ["POST"],
                "valid_params": [
                    "excludeFields",
                    "includeFields",
                    "lastModifiedDate",
                    "lastSeenAssetId",
                    "pageSize",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "token",
            },
            "get_asset": {
                "endpoint": "/am/v1/asset/host/id",
                "method": ["POST"],
                "valid_params": ["excludeFields", "includeFields", "assetId"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "token",
            },
            "query_assets": {
                "endpoint": "/am/v1/assets/host/filter/list",
                "method": ["POST"],
                "valid_params": [
                    "filter",
                    "excludeFields",
                    "includeFields",
                    "lastModifiedDate",
                    "lastSeenAssetId",
                    "pageSize",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "token",
            },
        },
    }
)



"""
Asset data class
"""

from dataclasses import dataclass, asdict
from typing import Union

from ...base.base_list import BaseList

"""
TODO: Move BaseCls into the package's base module
so it can be used universally. Things you realize 
months into a project...
"""


@dataclass
class BaseCls:
    pass

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)


@dataclass
class hostInstance(BaseCls):
    id: int = None
    port: int = None
    fqdn: str = None
    protocol: str = None
    service: str = None
    grade: str = None

    def __str__(self):
        return str(getattr(self, "id", ""))

    def __int__(self):
        return int(getattr(self, "id", 0))


@dataclass
class assetInterface(BaseCls):
    hostname: str = None
    address: str = None

    def __str__(self):
        return self.hostname if self.hostname else self.address


@dataclass
class Tag(BaseCls):
    name: str = None
    uuid: str = None

    def __str__(self):
        return self.name if self.name else self.uuid


@dataclass
class Asset(BaseCls):
    """
    Represents an asset in CERT,
    underneath a certificate
    data class.
    """

    id: int = None
    uuid: str = None
    netbiosName: str = None
    name: str = None
    operatingSystem: str = None
    hostInstances: Union[list[dict], BaseList[object]] = None
    assetInterfaces: Union[list[dict], BaseList[object]] = None
    tags: Union[list[str], BaseList[str]] = None
    primaryIp: str = None

    def __post_init__(self):
        if self.hostInstances:
            setattr(
                self,
                "hostInstances",
                BaseList([hostInstance(**x) for x in self.hostInstances]),
            )

        if self.assetInterfaces:
            setattr(
                self,
                "assetInterfaces",
                BaseList([assetInterface(**x) for x in self.assetInterfaces]),
            )

        if self.tags:
            setattr(self, "tags", BaseList([Tag(**x) for x in self.tags]))

    def __str__(self):
        return getattr(self, "name", "")



"""
Web Application Scanning call schema
"""

from frozendict import frozendict

WAS_SCHEMA = frozendict(
    {
        "was": {
            "url_type": "api",
            "call_webapp_api": {
                "endpoint": "/qps/rest/3.0/{placeholder}/was/webapp/{webappId}",
                "method": ["POST", "GET"],
                "valid_params": ["placeholder", "webappId", "_xml_data", "action"],
                "valid_POST_data": [
                    "_xml_data",
                    "id",
                    "name",
                    "url",
                    "tags.name",
                    "tags.id",
                    "createdDate",
                    "updatedDate",
                    "isScheduled",
                    "isScanned",
                    "lastScan.status",
                    "lastScan.date",
                    "verbose",
                ],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": True,
                "auth_type": "basic",
                "_xml_data": True,
            },
            "call_auth_api": {
                "endpoint": "/qps/rest/3.0/{placeholder}/was/webappauthrecord/{webappAuthRecordId}",
                "method": ["POST", "GET"],
                "valid_params": [
                    "placeholder",
                    "webappAuthRecordId",
                    "_xml_data",
                    "action",
                ],
                "valid_POST_data": [
                    "_xml_data",
                    "id",
                    "name",
                    "tags",
                    "tags.name",
                    "tags.id",
                    "createdDate",
                    "updatedDate",
                    "lastScan.date",
                    "lastScan.authStatus",
                    "isUsed",
                    "contents",
                ],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": True,
                "auth_type": "basic",
                "_xml_data": True,
            },
            "call_findings_api": {
                "endpoint": "/qps/rest/3.0/{placeholder}/was/finding/{findingId}",
                "method": ["POST", "GET"],
                "valid_params": ["placeholder", "findingId", "_xml_data", "action"],
                "valid_POST_data": [
                    "_xml_data",
                    "id",
                    "uniqueId",
                    "qid",
                    "name",
                    "type",
                    "url",
                    "webApp.tags.id",
                    "webApp.tags.name",
                    "status",
                    "patch",
                    "webApp.id",
                    "webApp.name",
                    "severity",
                    "externalRef",
                    "ignoredDate",
                    "ignoredReason",
                    "group",
                    "owasp.name",
                    "owasp.code",
                    "cwe.id",
                    "firstDetectedDate",
                    "lastDetectedDate",
                    "lastTestedDate",
                    "timesDetected",
                    "severity level",
                ],
                "use_requests_json_data": False,
                "return_type": "xml",
                "pagination": True,
                "auth_type": "basic",
                "_xml_data": True,
            },
        },
    }
)



"""
Contains the various modules' call schemas. These
files are joined together in ../call_schema.py
"""

from .cloudagent_schema import CLOUDAGENT_SCHEMA
from .cs_schema import CS_SCHEMA
from .gav_schema import GAV_SCHEMA
from .pm_schema import PM_SCHEMA
from .totalcloud_schema import TOTALCLOUD_SCHEMA
from .vmdr_schema import VMDR_SCHEMA
from .was_schema import WAS_SCHEMA
from .certview_schema import CERTVIEW_SCHEMA



"""
Checks that the pageSize value is not greater than a defined limit.
"""


def check_page_size_limit(page_size: int, limit=10_000) -> None:
    if not isinstance(page_size, (int, float)) or page_size > limit or page_size < 1:
        raise ValueError(
            f"pageSize must be an integer greater than 0 and less than or equal to {limit}"
        )



"""
Contains the Tag class.
"""

from dataclasses import dataclass, asdict


@dataclass
class Tag:
    """
    Represents a Tag in Patch Management.
    """

    name: str = None
    id: str = None

    def to_dict(self):
        """
        Convert the object to a dictionary.
        """
        return asdict(self)

    def __str__(self) -> str:
        return self.name if self.name else self.id

    @classmethod
    def from_dict(cls, data: dict):
        """
        Create a Tag object from a dictionary.
        """
        # exclusion tags only return tag ID as a string
        if isinstance(data, str):
            return cls(id=data)
        else:
            return cls(**data)



"""
Contains code to interact with Patch
Management jobs in Qualys.
"""

from __future__ import annotations
from typing import Union, Literal, overload, Sequence
from json import JSONDecodeError
from threading import Thread, Lock, current_thread
from queue import Queue

from .data_classes import *
from .base.page_limit import check_page_size_limit
from ..base.call_api import call_api
from ..base.base_list import BaseList
from ..auth.token import TokenAuth
from ..exceptions.Exceptions import *


def manage_jobs(
    auth: TokenAuth,
    method: Literal["GET", "POST", "PATCH", "DELETE"] = "GET",
    _use_singular_in_url: bool = False,
    **kwargs,
) -> dict:
    """
    Backend function to interact with PM jobs.
    This is not meant to be called directly.

    Args:
        auth (TokenAuth): The authentication object.
        method (Literal['GET','POST','PATCH','DELETE']): The HTTP method to use. Default is 'GET'.
        _use_singular_in_url (bool): Used for jobs and reports. URLs can be */job/* or */jobs/*.

    ## Kwargs:

        - kwargs from the calling function.

    Returns:
        API Response
    """

    if kwargs.get("platform"):
        kwargs["platform"] = kwargs["platform"].title()

    if kwargs.get("platform") not in ["Windows", "Linux", None]:
        raise ValueError(
            f"Invalid platform {kwargs.get('platform')}. Valid platforms are 'windows' or 'linux'"
        )

    if method not in ["GET", "POST", "PATCH", "DELETE"]:
        raise ValueError(
            f"Invalid method {method}. Valid methods are 'GET', 'POST', 'PATCH', 'DELETE'"
        )

    if kwargs.get("pageSize"):
        check_page_size_limit(kwargs["pageSize"])

    call_data = None

    call_data = call_api(
        auth=auth,
        module="pm",
        endpoint="deploymentjob" if _use_singular_in_url else "deploymentjobs",
        params=kwargs if method in ["GET", "DELETE"] else {},
        payload=kwargs if method in ["POST", "PATCH"] else {},
        override_method=method,
    )

    if call_data.status_code not in range(200, 299):
        err_data = call_data.json()

        # Check for 2501 error code - indicating we have hit the
        # limit of resources we can pull. (Fix this, Qualys!!!)
        if err_data["_error"].get("errorCode") == "2501":
            raise QualysAPIError(call_data.json())

    return call_data


def _list_jobs_backend(
    auth: TokenAuth,
    platform: Literal["Windows", "Linux"],
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> BaseList[PMJob]:
    """
    Backend function for listing jobs.

    Not meant to be called directly.

    **_result_bl is a shared resource for threads.
    """
    if page_count != "all":
        if not isinstance(page_count, int) or page_count < 1:
            raise ValueError(
                "page_count must be an integer greater than or equal to 1, or 'all'"
            )

    kwargs["platform"] = platform
    pages_pulled = 0
    # If called from a thread, we need to use a shared resource
    responses = (
        BaseList()
        if "_response_bl" not in kwargs.keys()
        else kwargs.pop("_response_bl")
    )

    # Set up URL format
    kwargs["placeholder"] = "summary"
    kwargs["pageNumber"] = pages_pulled
    lock = Lock()

    while True:
        response = manage_jobs(auth=auth, **kwargs)

        if not response:
            break

        try:
            parsed = response.json()
        except JSONDecodeError:
            raise QualysAPIError(response.status_code, response.text)

        if len(parsed) < 1:
            break

        # Protect BaseList if threads are used:
        with lock:
            responses.extend([PMJob.from_dict(**job) for job in parsed])

        if page_count != "all":
            pages_pulled += 1
            kwargs["pageNumber"] = pages_pulled
            if pages_pulled >= page_count:
                print(
                    f"{current_thread().name} - Hit user-defined limit of {page_count} pages for platform={platform}."
                )
                break
        else:
            pages_pulled += 1
            kwargs["pageNumber"] = pages_pulled

        if len(parsed) == 0:
            break

    return responses


def list_jobs(
    auth: TokenAuth,
    platform: Literal["all", "windows", "linux"] = "all",
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> BaseList[PMJob]:
    """
    Return a list of Patch Management jobs
    in the user's scope.

    Args:
        auth (TokenAuth): The authentication object.
        platform (Literal['all', 'windows', 'linux']): The platform to get jobs for. Default is 'all'.
        page_count (int): The number of pages to return. Default is 'all'.

    ## Kwargs:

        - filter (str): a Windows/Linux Job QQL filter.
        - attributes (str): The job attributes to return as a comma-separated string.
        - coauthorJob (bool): Only return jobs that the user is a co-author of (Default is False).
        - ownedJob (bool): Only return jobs that the user solely owns (Default is False).
        - pageSize (int): The number of jobs to return per page. Default is 10.

    Returns:
        BaseList: The response from the API as a BaseList of Job objects.
    """

    match platform.title():
        case "All":
            # Shared resource for threads to return results
            kwargs["_response_bl"] = BaseList()
            threads = [
                Thread(
                    target=_list_jobs_backend,
                    args=(auth, "Windows", page_count),
                    kwargs=kwargs,
                    name="WindowsJobThread",
                ),
                Thread(
                    target=_list_jobs_backend,
                    args=(auth, "Linux", page_count),
                    kwargs=kwargs,
                    name="LinuxJobThread",
                ),
            ]

            print(f"Spawned threads for both Windows and Linux jobs...")

            for thread in threads:
                thread.start()

            for thread in threads:
                thread.join()

            results = kwargs["_response_bl"]

        case "Windows":
            results = _list_jobs_backend(
                auth=auth, platform="Windows", page_count=page_count, **kwargs
            )
        case "Linux":
            results = _list_jobs_backend(
                auth=auth, platform="Linux", page_count=page_count, **kwargs
            )
        case _:
            raise ValueError(
                f"Invalid platform {platform}. Valid platforms are 'all', 'windows' or 'linux'"
            )

    return results


# Overload 1 for str jobId
@overload
def get_job_results(auth: TokenAuth, jobId: str, **kwargs) -> JobResultSummary:
    ...


# Overload 2 for list/BaseList of PMJob
@overload
def get_job_results(
    auth: TokenAuth, jobId: Union[list[PMJob], BaseList[PMJob]], **kwargs
) -> BaseList[JobResultSummary]:
    ...


def get_job_results(auth: TokenAuth, jobId: Union[str, Sequence[PMJob]], **kwargs):
    """
    Returns the results of a Patch Management job, or
    a specific instance of a job.

    Args:
        auth (TokenAuth): The authentication object.
        jobId (Union[str, BaseList[PMJob]]): The ID of the job to get results for.
        If a list of PMJob objects is passed, the function will thread the requests.

    ## Kwargs:

        - jobInstanceId (str): The ID of a specific instance of a job.
        - pageSize (int): The number of results to return per page. Default is 10.
        - sort (str): The field to sort results by.

    Returns:
        Union[JobResultSummary, BaseList[JobResultSummary]]: The response from the
        API as a JobResultSummary object or a BaseList of JobResultSummary objects.
    """
    lock = Lock()
    if isinstance(jobId, str):
        kwargs["placeholder"] = f"{jobId}/deploymentjobresult/summary"
        kwargs_no_bl = kwargs.copy()
        kwargs_no_bl.pop("_response_bl", None)
        result = manage_jobs(
            auth=auth, method="POST", _use_singular_in_url=True, **kwargs_no_bl
        )
        if not kwargs.get("_response_bl") and kwargs.get("_response_bl") != BaseList():
            return JobResultSummary.from_dict(**result.json())
        else:
            with lock:
                kwargs["_response_bl"].extend(
                    [JobResultSummary.from_dict(**result.json())]
                )
            return

    elif isinstance(jobId, (list, BaseList)) and all(
        isinstance(job, (PMJob, str)) for job in jobId
    ):
        # Shared resource for threads to return results
        kwargs["_response_bl"] = BaseList()
        threads = []
        print(f"Spawning threads for {len(jobId)} jobs...")
        q = Queue()
        for job in jobId:
            if isinstance(job, PMJob):
                q.put(job.id)
            else:
                q.put(job)

        # Each thread will pop a job from the queue and process it.
        # at no point should the threads list have a len > 5.
        while not q.empty():
            if len(threads) < 5:
                job = q.get()
                thread = Thread(
                    target=get_job_results,
                    args=(auth, job),
                    kwargs=kwargs,
                    name=f"JobResultThread-{job}",
                )
                thread.start()
                threads.append(thread)
            else:
                for thread in threads:
                    thread.join()
                threads = []

            with lock:
                if q.qsize() % 25 == 0:
                    print(f"Requests left in queue: {q.qsize()}")

        for thread in threads:
            thread.join()

        return kwargs["_response_bl"]

    else:
        raise ValueError(
            "jobId must be a string, or a list/BaseList of PMJob objects or strings."
        )


def get_job_runs(
    auth: TokenAuth, jobId: Union[str, BaseList[PMJob], list[PMJob]]
) -> BaseList[PMRun]:
    """
    Get a list of job runs for a specific job.

    Args:
        auth (TokenAuth): The authentication object.
        jobId (str) The ID of the job to get runs for.

    Returns:
        BaseList[PMRun]: The response from the API as a BaseList of PMRun objects.
    """

    params = {"placeholder": f"{jobId}/runs"}

    result = manage_jobs(auth=auth, _use_singular_in_url=True, **params)

    result = result.json()

    for run in result:
        run["jobId"] = jobId

    return BaseList([PMRun.from_dict(**run) for run in result])



"""
Contains the functions to upload supported Patch Management API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame, concat
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_pm_jobs(
    jobs: BaseList,
    cnxn: Connection,
    table_name: str = "pm_jobs",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```pm.list_jobs```
    to a SQL database.

    Args:
        jobs (BaseList): A BaseList of PMJob objects.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to "pm_jobs".
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "coAuthorUserIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subCategory": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dynamicQQLType": types.Integer(),
        "matchAllTags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "type": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "recurringWeekDays": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "totalAssetCount": types.Integer(),
        "isDynamicPatchesQQL": types.Boolean(),
        "assetCount": types.Integer(),
        "customPatchUrlConfigured": types.Boolean(),
        "linkedJobReferenceCount": types.Integer(),
        "created_date": types.DateTime(),
        "created_user_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "created_user_id": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "triggerStatus": types.Integer(),
        "recurringDayOfMonth": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "jobSource": types.Integer(),
        "readOnly": types.Boolean(),
        "tags": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "exclusionAssetIds": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "scheduleType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dayOfMonth": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "disabledPatches": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "filterType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updated_date": types.DateTime(),
        "updated_user_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updated_user_id": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "additionalQQLS": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "timezone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "exclusionFilterType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "nextScheduleDateTime": types.DateTime(),
        "platform": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "mitigationActionCount": types.Integer(),
        "isPriorityJob": types.Boolean(),
        "customerId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "exclusionTags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "taggedAssetCount": types.Integer(),
        "patchCount": types.Integer(),
        "applicableAssetCount": types.Integer(),
        "patchTuesdayPlusXDays": types.Integer(),
        "schemaVersion": types.Integer(),
        "recurringLastDayOfMonth": types.Boolean(),
        "recurringWeekDayOfMonth": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "isRecurring": types.Boolean(),
        "lastScheduleDateTime": types.DateTime(),
        "dynamicPatchesQQL": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "startDateTime": types.DateTime(),
        "timezoneType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "deleted": types.Boolean(),
        "monthlyRecurringType": types.Integer(),
        "completionPercent": types.Float(),
        "category": types.Integer(),
        "linkedJobId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "assetResultReceivedCount": types.Integer(),
        "isVulnContext": types.Boolean(),
        "isAssetImported": types.Boolean(),
        "remediationQids": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(job) for job in jobs])

    # Drop cols that are parsed out into other fields:
    df.drop(
        columns=[
            "created",
            "updated",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)


def upload_pm_job_results(
    job_results: BaseList,
    cnxn: Connection,
    jobs_table_name: str = "pm_job_results_jobResults",
    assets_table_name: str = "pm_job_results_assets",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```pm.get_job_results```
    to 2 DB tables: ```pm_job_results_jobResults```
    and ```pm_job_results_assets``` by default.

    Args:
        job_results (BaseList): A BaseList of JobResultSummary objects.
        cnxn (Connection): The Connection object to the SQL database.
        jobs_table_name (str): The name of the table to upload to. Defaults to "pm_job_results".
        assets_table_name (str): The name of the table to upload to. Defaults to "pm_job_results_assets".
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    """
    Separate job summary and assets list into two separate tables.
    We can link assets back to a job by putting the job.id in the assets table
    with pandas:
    """

    # Explicitly set the import datetime if not provided
    # so we can keep the two tables in sync:
    if not override_import_dt:
        override_import_dt = datetime.now()

    assets_df = DataFrame()
    job_results_df = DataFrame()

    for job_result in job_results:
        for asset in job_result.assets:
            asset_dict = prepare_dataclass(asset)
            assets_df = concat([assets_df, DataFrame([asset_dict])], ignore_index=True)

        job_results_df = concat(
            [job_results_df, DataFrame([prepare_dataclass(job_result)])],
            ignore_index=True,
        )

    # Upload the data:

    COLS = {
        "id": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "assetCount": types.Integer(),
        "patchCount": types.Integer(),
        "createdBy": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "createdOn": types.DateTime(),
        "assets": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
    }

    job_summaries_uploaded = upload_data(
        job_results_df, jobs_table_name, cnxn, COLS, override_import_dt
    )

    print(
        f"Uploaded {job_summaries_uploaded} to {jobs_table_name}. Moving to assets..."
    )

    COLS = {
        "id": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "operatingSystem": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "jobId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "tags": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "lastLoggedOnUser": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "successPatches": types.Integer(),
        "installedPatches": types.Integer(),
        "failedPatches": types.Integer(),
        "supersededPatches": types.Integer(),
        "notApplicablePatches": types.Integer(),
        "executing": types.Boolean(),
        "pendingExecution": types.Boolean(),
        "pendingReboot": types.Boolean(),
        "pendingVerification": types.Boolean(),
        "jobInstanceId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "interfaces": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "skipPatchCount": types.Integer(),
        "additionalFields": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "endDateTime": types.DateTime(),
        "startDateTime": types.DateTime(),
        "statusDateTime": types.DateTime(),
        "status": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "statusCode": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "jobSentOn": types.DateTime(),
        "installed": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "failed": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "success": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "superseded": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "notApplicable": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "failedActionsCount": types.Integer(),
        "successfulActionsCount": types.Integer(),
        "skippedActionsCount": types.Integer(),
        "interimResultStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "totalPatchCount": types.Integer(),
        "runId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "scanDateTime": types.DateTime(),
        "pendingForRebootInAnotherJob": types.Boolean(),
        "pendingForRebootInAnotherJobName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "osIdentifier": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    assets_uploaded = upload_data(
        assets_df, assets_table_name, cnxn, COLS, override_import_dt
    )

    return assets_uploaded


def upload_pm_job_runs(
    runs: BaseList,
    cnxn: Connection,
    table_name: str = "pm_job_runs",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```pm.get_job_runs```
    to a SQL database.

    Args:
        runs (BaseList): A BaseList of PMJobRun objects.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to "pm_job_runs".
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "jobInstanceId": types.Integer(),
        "jobId": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "scheduledDateTime": types.DateTime(),
        "timezoneType": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(run) for run in runs])

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)



"""
Contains the functions to upload supported Certificate View API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame, concat
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_cert_certs(
    certs: BaseList,
    cnxn: Connection,
    certs_table_name: str = "cert_certs",
    assets_table_name: str = "cert_assets",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```cert.list_certs```
    to 2 DB tables: ```cert_certs``` and ```cert_assets``` by default.

    Args:
        certs (BaseList): A BaseList of Certificate objects.
        cnxn (Connection): The Connection object to the SQL database.
        certs_table_name (str): The name of the table to upload to. Defaults to "cert_certs".
        assets_table_name (str): The name of the table to upload to. Defaults to "cert_assets".
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    """
    Separate certs and assets list into two separate tables.
    We can link assets back to a cert by putting the cert.id in the assets table
    with pandas:
    """

    # Explicitly set the import datetime if not provided
    # so we can keep the two tables in sync:
    if not override_import_dt:
        override_import_dt = datetime.now()

    assets_df = DataFrame()
    certs_df = DataFrame()

    for cert in certs:
        for asset in cert.assets:
            asset_dict = prepare_dataclass(asset)
            # Add the cert ID to the asset so we can link them:
            asset_dict["certId"] = cert.id
            assets_df = concat([assets_df, DataFrame([asset_dict])], ignore_index=True)

        certs_df = concat(
            [certs_df, DataFrame([prepare_dataclass(cert)])],
            ignore_index=True,
        )

    # Upload the data:

    COLS = {
        "id": types.Integer(),
        "certhash": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "keySize": types.Integer(),
        "serialNumber": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "validToDate": types.DateTime(),
        "validTo": types.DateTime(),
        "validFromDate": types.DateTime(),
        "validFrom": types.DateTime(),
        "signatureAlgorithm": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "extendedValidation": types.Boolean(),
        "createdDate": types.DateTime(),
        "dn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "subject_organization": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subject_locality": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subject_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subject_country": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "subject_organizationUnit": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "updateDate": types.DateTime(),
        "lastFound": types.DateTime(),
        "imported": types.Boolean(),
        "selfSigned": types.Boolean(),
        "issuer_organization": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuer_organizationUnit": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuer_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuer_country": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuer_state": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuer_certhash": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuer_locality": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "issuerCategory": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "instanceCount": types.Integer(),
        "assetCount": types.Integer(),
        "sources": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "assets": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "type": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "rootissuer_organization": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rootissuer_organizationUnit": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rootissuer_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rootissuer_country": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rootissuer_state": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rootissuer_certhash": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "rootissuer_locality": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Drop columns that we parsed out:
    certs_df.drop(
        columns=[
            "subject",
            "issuer",
            "rootissuer",
        ],
        inplace=True,
    )

    certs_uploaded = upload_data(
        certs_df, certs_table_name, cnxn, COLS, override_import_dt
    )

    print(f"Uploaded {certs_uploaded} to {certs_table_name}. Moving to assets...")

    COLS = {
        "id": types.Integer(),
        "certId": types.Integer(),
        "uuid": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "netbiosName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "operatingSystem": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "hostInstances": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "assetInterfaces": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "tags": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "primaryIp": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    assets_uploaded = upload_data(
        assets_df, assets_table_name, cnxn, COLS, override_import_dt
    )

    print(f"Uploaded {assets_uploaded} to {assets_table_name}.")

    return certs_uploaded



"""
Contains the list_certs function
"""

from json import JSONDecodeError
from typing import Union

from .data_classes.Certificate import Certificate
from ..base.base_list import BaseList
from ..base.call_api import call_api
from ..auth.token import TokenAuth
from ..exceptions.Exceptions import QualysAPIError


def list_certs(
    auth: TokenAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> dict:
    """
    Return a list of certificates that match the given kwargs.

    Args:
        auth (TokenAuth): The authentication object for the API
        page_count (Union[int, 'all']): The number of pages to retrieve. If 'all', retrieve all pages. Default is 'all'.
        **kwargs: The search parameters for the API call.

    **Kwargs:

        - certId (str): The certificate ID.
        - certId_operator (Literal["IN", "LESSER", "IS_EMPTY", "GREATER", "GREATER_THAN_EQUAL", "IS_NOT_EMPTY", "EQUALS", "NOT_EQUALS", "LESS_THAN_EQUAL", "CONTAINS"]): The operator to use for the certId parameter. Default is "IS_NOT_EMPTY".
        - hash (str): The certificate hash.
        - hash_operator (Literal["IN", "LESSER", "IS_EMPTY", "GREATER", "GREATER_THAN_EQUAL", "IS_NOT_EMPTY", "EQUALS", "NOT_EQUALS", "LESS_THAN_EQUAL", "CONTAINS"]): The operator to use for the hash parameter. Default is "IS_NOT_EMPTY".
        - validFromDate (str): The valid from date.
        - validFromDate_operator (Literal["IN", "LESSER", "IS_EMPTY", "GREATER", "GREATER_THAN_EQUAL", "IS_NOT_EMPTY", "EQUALS", "NOT_EQUALS", "LESS_THAN_EQUAL", "CONTAINS"]): The operator to use for the validFromDate parameter. Default is "GREATER".
        - wasUrl (str): The WAS URL.
        - wasUrl_operator (Literal["IN", "LESSER", "IS_EMPTY", "GREATER", "GREATER_THAN_EQUAL", "IS_NOT_EMPTY", "EQUALS", "NOT_EQUALS", "LESS_THAN_EQUAL", "CONTAINS"]): The operator to use for the wasUrl parameter. Default is "IS_NOT_EMPTY".
        - certificateType (Literal["Leaf", "Intermediate", "Root"]): The certificate type.
        - pageSize (int): The number of items to return per page. Default is 10.

    Returns:
        dict: The JSON response from the API.
    """

    # Check that page_count is either > 0 or 'all':
    if page_count != "all" and (not isinstance(page_count, int) or page_count <= 0):
        raise ValueError("page_count must be 'all' or > 0")

    pages_pulled = 0
    responses = BaseList()

    payload = {
        "filter": {
            "filters": [],
            "operator": "AND",
        },
        "pageNumber": 0,
        "pageSize": kwargs.pop("pageSize", 10),
        "includes": ["ASSET_INTERFACES", "ASSET_TAGS"],
        "assetType": "ALL",
    }

    """
    Filter structure:
    
    filters: [
        {
            "value": "string",
            "field": "string",
            "operator": "string"
        },
        ...
    ]
    """

    # Set up the filters from the kwargs.
    # Group by which _operator Literal is valid for each parameter.

    for key, value in kwargs.items():
        if key.endswith("_operator"):
            continue
        operator = (
            kwargs.get(f"{key}_operator", "IS_NOT_EMPTY")
            if key != "validFromDate"
            else "GREATER"
        )
        if operator not in [
            "IN",
            "LESSER",
            "IS_EMPTY",
            "GREATER",
            "GREATER_THAN_EQUAL",
            "IS_NOT_EMPTY",
            "EQUALS",
            "NOT_EQUALS",
            "LESS_THAN_EQUAL",
            "CONTAINS",
        ]:
            raise ValueError(f"Invalid operator: {operator}")

        if key == "certificateType":
            payload["filter"]["filters"].append(
                {"field": "certificate.type", "value": value, "operator": "EQUALS"}
            )
        else:
            payload["filter"]["filters"].append(
                {"field": key, "value": value, "operator": operator}
            )

    while True:
        response = call_api(
            auth=auth, module="cert", endpoint="list_certs", jsonbody=payload
        )

        response_json = response.json()
        if response.status_code != 200:
            raise QualysAPIError(response.text)

        for cert in response_json:
            responses.append(Certificate(**cert))
        pages_pulled += 1

        if page_count != "all" and pages_pulled >= page_count:
            print(f"Hit user-defined limit of {page_count} pages.")
            break

        if not response_json:
            break

        payload["pageNumber"] += 1

    return responses



"""
certificate view call schemas
"""

from frozendict import frozendict

CERTVIEW_SCHEMA = frozendict(
    {
        "cert": {
            "url_type": "gateway",
            "list_certs": {
                "endpoint": "/certview/v2/certificates",
                "method": ["POST"],
                "valid_params": [],
                "valid_POST_data": [
                    "certId",
                    "pageSize",
                    "pageNumber",
                    "hash",
                    "validFromDate",
                    "wasUrl",
                    "certificateType",
                ],
                "use_requests_json_data": True,
                "return_type": "json",
                "pagination": True,
                "auth_type": "token",
            },
        },
    }
)



"""
TotalCloud call schema
"""

from frozendict import frozendict

TOTALCLOUD_SCHEMA = frozendict(
    {
        "cloudview": {  # TOTALCLOUD
            "url_type": "base",
            "get_connectors": {
                "endpoint": "/cloudview-api/rest/v1/{cloudprovider}/connectors",
                "method": ["GET"],
                "valid_params": [
                    "pageNo",
                    "pageSize",
                    "filter",
                    "sort",
                    "cloudprovider",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "basic",
            },
            "get_connector_details": {
                "endpoint": "/cloudview-api/rest/v1/{cloudprovider}/connectors/{placeholder}",
                "method": ["GET"],
                "valid_params": ["placeholder", "cloudprovider"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "basic",
            },
            "get_aws_base_account": {
                "endpoint": "/cloudview-api/rest/v1/aws/connectors/awsBaseAccountId",
                "method": ["GET"],
                "valid_params": [],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "basic",
            },
            "get_control_metadata": {
                "endpoint": "/cloudview-api/rest/v1/controls/metadata/list",
                "method": ["GET"],
                "valid_params": ["pageNo", "pageSize", "filter"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "basic",
            },
            "get_control_stats_by_resouce": {
                "endpoint": "/cloudview-api/rest/v1/{cloudprovider}/evaluations/stats/{controlid}/{connectorid}",
                "method": ["GET"],
                "valid_params": [
                    "controlid",
                    "connectorid",
                    "cloudprovider",
                    "resourceId",
                    "provider",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "basic",
            },
            "get_inventory": {
                "endpoint": "/cloudview-api/rest/v1/resource/{placeholder}/{cloudprovider}",
                "method": ["GET"],
                "valid_params": [
                    "cloudprovider",
                    "placeholder",
                    "filter",
                    "pageNo",
                    "pageSize",
                    "sort",
                    "updated",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "basic",
            },
            "get_resource_details": {
                "endpoint": "/cloudview-api/rest/v1/resource/{placeholder}/uuid/{resourceid}/{cloudprovider}",
                "method": ["GET"],
                "valid_params": [
                    "cloudprovider",
                    "placeholder",
                    "resourceid",
                    "pageNo",
                    "pageSize",
                    "filter",
                    "sort",
                    "updated",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "basic",
            },
            "get_account_evaluation": {
                "endpoint": "/cloudview-api/rest/v1/{cloudprovider}/evaluations/{placeholder}",
                "method": ["GET"],
                "valid_params": ["filter", "placeholder", "cloudprovider", "accountId"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "basic",
            },
            "get_resources_evaluated_by_control": {
                "endpoint": "/cloudview-api/rest/v1/{cloudprovider}/evaluations/{placeholder}/resources/{controlid}",
                "method": ["GET"],
                "valid_params": [
                    "accountId",  # placeholder ⬇️
                    "controlid",
                    "filter",
                    "placeholder",
                    "cloudprovider",
                ],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": False,
                "auth_type": "basic",
            },
            "get_remediation_activities": {
                "endpoint": "/cloudview-api/rest/v1/remediation/activity",
                "method": ["GET"],
                "valid_params": ["pageNo", "pageSize", "filter", "cloudType"],
                "valid_POST_data": [],
                "use_requests_json_data": False,
                "return_type": "json",
                "pagination": True,
                "auth_type": "basic",
            },
        },
    }
)
